"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2021/9/3 下午11:49
"""
import traceback
import logging
import pymysql
from scrapy.exceptions import NotConfigured
from twisted.enterprise import adbapi
from twisted.internet import defer

logger = logging.getLogger(__name__)


class MysqlBasePipeline(object):
    def __init__(self, host, port, user, password, db, charset, table, update_fixed_fields):
        params = dict(
            host=host,
            port=port,
            user=user,
            password=password,
            db=db,
            charset=charset,
            cursorclass=pymysql.cursors.DictCursor
        )
        self.table = table
        self.update_fixed_fields = update_fixed_fields
        self.dbpool = adbapi.ConnectionPool('pymysql', **params)

    @classmethod
    def from_crawler(cls, crawler):
        host = crawler.settings.get('MYSQL_HOST')
        port = crawler.settings.getint('MYSQL_PORT', 3306)
        user = crawler.settings.get('MYSQL_USERNAME')
        password = crawler.settings.get('MYSQL_PASSWORD')
        db = crawler.settings.get('MYSQL_DB')
        charset = crawler.settings.get('MYSQL_CHARSET', 'utf8')
        table = crawler.settings.get('MYSQL_TABLE')
        update_fixed_fields = crawler.settings.get('MYSQL_UPDATE_FIXED_FIELDS')

        if host is None or db is None or table is None:
            raise NotConfigured

        return cls(host, port, user, password, db, charset, table, update_fixed_fields)

    @defer.inlineCallbacks
    def process_item(self, item, spider):
        try:
            yield self.dbpool.runInteraction(self.do_mysql, dict(item))
        except:
            logger.error(traceback.format_exc())

        defer.returnValue(item)

    def do_mysql(self, curson, item):
        """Insert operation or update operation"""


class MysqlInsertPipeline(MysqlBasePipeline):
    def do_mysql(self, cursor, item):
        fields = ','.join(item.keys())
        values = ','.join(('%({})s'.format(k) for k in item.keys()))
        sql = f'insert ignore into {self.table}({fields}) values({values})'
        cursor.execute(sql, item)


class MysqlUpdatePipeline(MysqlBasePipeline):
    def do_mysql(self, cursor, item):
        update_phrase = [f'{name}=%({name})s' for name in item.keys() if name not in self.update_fixed_fields]
        where_phrase = [f'{name}=%({name})s' for name in self.update_fixed_fields]
        update_sql = f'update {self.table} set {",".join(update_phrase)} where {" and ".join(where_phrase)}'
        cursor.execute(update_sql, item)




from scrapy_box.middlewares.proxymiddleware import \
    RandomProxyDownloaderMiddleware, \
    SplashProxyDownloaderMiddleware, \
    SplashRetryProxyMiddleware

from scrapy_box.middlewares.retrymiddleware import ErrorRedirectMiddleware, ErrorResponseMiddleware
from scrapy_box.pipelines.mongo import MongoBatchInsertPipeline
from scrapy_box.pipelines.mysql import MysqlInsertPipeline, MysqlUpdatePipeline



"""
@Description:
Default downloader middlewares and orders
{
    'scrapy.downloadermiddlewares.robotstxt.RobotsTxtMiddleware': 100,
    'scrapy.downloadermiddlewares.httpauth.HttpAuthMiddleware': 300,
    'scrapy.downloadermiddlewares.downloadtimeout.DownloadTimeoutMiddleware': 350,
    'scrapy.downloadermiddlewares.defaultheaders.DefaultHeadersMiddleware': 400,
    'scrapy.downloadermiddlewares.useragent.UserAgentMiddleware': 500,
    'scrapy.downloadermiddlewares.retry.RetryMiddleware': 550,
    'scrapy.downloadermiddlewares.ajaxcrawl.AjaxCrawlMiddleware': 560,
    'scrapy.downloadermiddlewares.redirect.MetaRefreshMiddleware': 580,
    'scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware': 590,
    'scrapy.downloadermiddlewares.redirect.RedirectMiddleware': 600,
    'scrapy.downloadermiddlewares.cookies.CookiesMiddleware': 700,
    'scrapy.downloadermiddlewares.httpproxy.HttpProxyMiddleware': 750,
    'scrapy.downloadermiddlewares.stats.DownloaderStats': 850,
    'scrapy.downloadermiddlewares.httpcache.HttpCacheMiddleware': 900,
}

@Usage: 
@Author: liuxianglong
@Date: 2021/9/1 下午8:15
"""



"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2021/9/2 下午4:12
"""
import logging
from pymongo import MongoClient
from scrapy.exceptions import NotConfigured

logger = logging.getLogger(__name__)


class MongoBatchInsertPipeline:
    def __init__(self, mongo_client, mongo_db, mongo_collection, mongo_insert_size):
        self.data = list()
        self.insert_size = mongo_insert_size
        self.mongo_cli = mongo_client
        self.mongo_table = self.mongo_cli[mongo_db][mongo_collection]

    @classmethod
    def from_crawler(cls, crawler):
        addr = crawler.settings.get('MONGO_ADDR')
        username = crawler.settings.get('MONGO_USERNAME')
        password = crawler.settings.get('MONGO_PASSWORD')
        db = crawler.settings.get('MONGO_DB')
        collection = crawler.settings.get('MONGO_COLLECTION')
        insert_size = crawler.settings.get('MONGO_INSERT_SIZE', 100)

        if db is None or collection is None or addr is None:
            raise NotConfigured

        client = MongoClient(addr)
        client.admin.authenticate(username, password)

        return cls(client, db, collection, insert_size)

    def _process_item(self, item):
        """return item or None"""
        return dict(item)

    def process_item(self, item, spider):
        """ 批量插入数据库 """
        if self._process_item(item):
            self.data.append(item)

        if len(self.data) >= self.insert_size:
            logger.info(f'批量插入 {len(self.data)}条 数据到MongoDB')
            self.mongo_table.insert_many(self.data)
            logger.info('批量插入成功')
            self.data.clear()

        return item

    def close_spider(self, spider):
        """ 关闭爬虫，处理最后的数据 """
        if len(self.data) > 0:
            logger.info(f'批量插入 {len(self.data)}条 数据到MongoDB')
            self.mongo_table.insert_many(self.data)
            logger.info('批量插入成功')
            self.data.clear()

        self.mongo_cli.close()




"""
@Description:
@Usage:
@Author: liuxianglong
@Date: 2021/9/1 下午8:20
"""
import re
import time
import logging
import treq
from twisted.internet import defer
from scrapy.exceptions import NotConfigured

logger = logging.getLogger(__name__)


class RandomProxyDownloaderMiddleware(object):
    def __init__(self, proxy_addr):
        self.proxy_addr = proxy_addr

    @classmethod
    def from_crawler(cls, crawler):
        proxy_addr = crawler.settings.get('PROXY_ADDR')
        if proxy_addr is None or not proxy_addr.startswith('http'):
            raise NotConfigured

        return cls(proxy_addr)

    @defer.inlineCallbacks
    def process_request(self, request, spider):
        proxy = yield self.get_proxy()
        logger.debug('using proxy %s for request %s' % (proxy, request.url))
        request.meta['proxy'] = 'http://%(ip)s:%(port)s' % proxy

    @defer.inlineCallbacks
    def get_proxy(self):
        """ Get proxy from your PROXY_ADDR, this function should be rewrite if it can't satisfy your need.

            It Should return json format proxy like:
            {
                'ip': '192.168.0.32',
                'port': 23,
                ...
            }
        """
        while True:
            response = yield treq.get(url=self.proxy_addr)
            proxy = yield response.json()
            if proxy:
                defer.returnValue(proxy)
            else:
                time.sleep(2)
                logger.info('未获取到代理，等待两秒，重新获取')


class SplashProxyDownloaderMiddleware(object):
    def __init__(self, proxy_addr, splash_url):
        self.proxy_addr = proxy_addr
        self.splash_url = splash_url

    @classmethod
    def from_crawler(cls, crawler):
        proxy_addr = crawler.settings.get('PROXY_ADDR')
        splash_url = crawler.settings.get('SPLASH_URL')
        if proxy_addr is None or splash_url is None or not proxy_addr.startswith('http'):
            raise NotConfigured

        return cls(proxy_addr, splash_url)

    @defer.inlineCallbacks
    def process_request(self, request, spider):
        proxy = yield self.get_proxy()
        if not request.url.startswith(self.splash_url):
            logger.debug('using proxy %s for request %s' % (proxy, request.meta['splash']['args']['url']))
            request.meta['splash']['args']['proxy'] = 'http://%(ip)s:%(port)s' % proxy

    @defer.inlineCallbacks
    def get_proxy(self):
        """ Get proxy from your PROXY_ADDR, this function should be rewrite if it can't satisfy your need.

            It Should return json format proxy like:
            {
                'ip': '192.168.0.32',
                'port': 23,
                ...
            }
        """
        while True:
            response = yield treq.get(url=self.proxy_addr)
            proxy = yield response.json()
            if proxy:
                defer.returnValue(proxy)
            else:
                time.sleep(2)
                logger.info('未获取到代理，等待两秒，重新获取')


class SplashRetryProxyMiddleware(SplashProxyDownloaderMiddleware):
    retry_processed = dict()

    @defer.inlineCallbacks
    def process_request(self, request, spider):
        retry_times = request.meta.get('retry_times')
        if retry_times:
            if self.retry_processed.get(retry_times) is None:
                self.retry_processed[retry_times] = True

                proxy = yield self.get_proxy()
                body = request.body.decode()
                old_proxy = ''.join(re.findall(r'"proxy": "(.*?)"', body))
                if old_proxy:
                    logger.info(f"{request.meta['splash']['args']['url']} retry times: {retry_times}, using new proxy: {proxy}")
                    new_body = body.replace(old_proxy, 'http://%(ip)s:%(port)s' % proxy)
                    new_request = request.replace(
                        body=new_body.encode(),
                    )
                    return new_request



"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2021/9/1 下午8:50
"""
import re
import logging
from w3lib.url import safe_url_string
from urllib.parse import urljoin, urlparse
from scrapy.exceptions import IgnoreRequest, NotConfigured
from scrapy.downloadermiddlewares.retry import RetryMiddleware
from scrapy.downloadermiddlewares.redirect import BaseRedirectMiddleware

logger = logging.getLogger(__name__)


class ErrorRedirectMiddleware(BaseRedirectMiddleware, RetryMiddleware):
    """ This build-in RedirectMiddleware like below
        'scrapy.downloadermiddlewares.redirect.RedirectMiddleware': 600,

        You should use ErrorRedirectMiddleware like below, make use the num larger than 600

        'scrapy_box.ErrorRedirectMiddleware': 601
    """
    def __init__(self, settings):
        super(ErrorRedirectMiddleware, self).__init__(settings)
        self.max_retry_times = settings.getint('RETRY_TIMES')
        self.priority_adjust = settings.getint('RETRY_PRIORITY_ADJUST')
        self.error_redirect_url_snippet = settings.get('ERROR_REDIRECT_URL_SNIPPET', [])

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def process_response(self, request, response, spider):
        if (
            request.meta.get('dont_redirect', False)
            or response.status in getattr(spider, 'handle_httpstatus_list', [])
            or response.status in request.meta.get('handle_httpstatus_list', [])
            or request.meta.get('handle_httpstatus_all', False)
        ):
            return response

        allowed_status = (301, 302, 303, 307, 308)
        if 'Location' not in response.headers or response.status not in allowed_status:
            return response

        location = safe_url_string(response.headers['Location'])
        if response.headers['Location'].startswith(b'//'):
            request_scheme = urlparse(request.url).scheme
            location = request_scheme + '://' + location.lstrip('/')

        redirected_url = urljoin(request.url, location)

        for snippet in self.error_redirect_url_snippet:
            if snippet in redirected_url:
                logger.info('Error redirect url <%s> via <%s>, retry..' % (redirected_url, request.url))
                req = self._retry(request, "error_redirect", spider)
                if req is not None:
                    return req
                else:
                    raise IgnoreRequest

        if response.status in (301, 307, 308) or request.method == 'HEAD':
            redirected = request.replace(url=redirected_url)
            return self._redirect(redirected, request, spider, response.status)

        redirected = self._redirect_request_using_get(request, redirected_url)
        return self._redirect(redirected, request, spider, response.status)


class ErrorResponseMiddleware(RetryMiddleware):
    def __init__(self, settings):
        super(ErrorResponseMiddleware, self).__init__(settings)
        self.error_response_snippet = settings.get('ERROR_RESPONSE_SNIPPET', [])

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def error_validation(self, res):
        for error in self.error_response_snippet:
            if re.search(error, res.text):
                return True, error
        return False, 'success'

    def process_response(self, request, response, spider):
        iserror, error = self.error_validation(response)
        if iserror:
            logger.info(f"error response: {error} from <{request.url}>, retry...")
            req = self._retry(request, "error_response", spider)
            if req is not None:
                return req
            else:
                raise IgnoreRequest
        return response



"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2021/8/21 下午6:44
"""
# !/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='scrapy_box',
    version='V1.4.0',
    description=(
        '简单易用的scrapy插件盒'
    ),
    long_description=open('README.rst').read(),
    author='xlomg',
    author_email='liu_xianglong@live.com',
    maintainer='xlomg',
    maintainer_email='liu_xianglong@live.com',
    license='BSD License',
    packages=['scrapy_box.middlewares', 'scrapy_box.pipelines'],
    py_modules=["scrapy_box.__init__"],
    platforms=["all"],
    url='https://github.com/xlomg/scrapy_box.git',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'scrapy', 'treq', 'pymongo', 'pandas', 'pymysql'
    ]
)